grammar Expression;
options 
{
	output=AST;
	ASTLabelType=CommonTree;
	language=CSharp3;
}

tokens
{
	PARAM;
	REFERENCE;
	RANGEREFERENCE;
	NEGATE;
	STRING;
}

@members {
List<Exception> exceptions = new List<Exception>();

public override void ReportError(RecognitionException e)
{
    exceptions.Add(e);
}

public bool HasError
{
	get { return exceptions.Count > 0; }
}

public string ErrorMessage
{
	get { return this.GetErrorMessage(exceptions[0] as RecognitionException, this.TokenNames); }
}

public string ErrorPosition
{
	get { return this.GetErrorHeader(exceptions[0] as RecognitionException); }
}
}

@header {
using System;
using System.Text;
}


expression
	: 	logicalExpression EOF!
	;
		
logicalExpression
	:	booleanAndExpression (OR^ booleanAndExpression )* 
	;

OR 	: 	'||' | 'or';
	
booleanAndExpression
	:	equalityExpression (AND^ equalityExpression)* 
	;

AND 	: 	'&&' | 'and';

equalityExpression
	:	relationalExpression ((EQUALS|NOTEQUALS)^ relationalExpression)*
	;

EQUALS	
	:	'=' | '==';
NOTEQUALS 
	:	'!=' | '<>';

relationalExpression
	:	additiveExpression ( (LT|LTEQ|GT|GTEQ)^ additiveExpression )*
	;

LT	:	'<';
LTEQ	:	'<=';
GT	:	'>';
GTEQ	:	'>=';

additiveExpression
	:	multiplicativeExpression ( (PLUS|MINUS)^ multiplicativeExpression )*
	;

PLUS	:	'+';
MINUS	:	'-';

multiplicativeExpression 
	:	powerExpression ( (MULT|DIV|MOD)^ powerExpression )*
	;
	
MULT	:	'*';
DIV	:	'/';
MOD	:	'%';

powerExpression 
	:	unaryExpression ( POW^ unaryExpression )*
	;
	
POW	:	'^';

unaryExpression
	:	primaryExpression
    	|	NOT^ primaryExpression
    	|	MINUS primaryExpression -> ^(NEGATE primaryExpression)
   	;
  
NOT	:	'!' | 'not';

primaryExpression
	:	'('! logicalExpression ')'!
	|	value 
	;

value	
	: 	INTEGER
	|	FLOAT
	|	STRING
	|	function
	|	reference
	|	rangereference
	|	par
	;

reference
	:	REF		-> ^(REFERENCE	REF?)
	;
	
rangereference
	:	REFRANGE	-> ^(RANGEREFERENCE	REFRANGE?)
	;
	
par	:	PAR		-> ^(PARAM	PAR?)
	;

PAR	:	('X' | 'Y')
	;

REF	:	('A'..'Z')+ ('0'..'9')+;

REFRANGE
	:	(REF ':' REF)
	;

STRING
    	:  	'\'' ( EscapeSequence | (options {greedy=false;} : ~('\u0000'..'\u001f' | '\\' | '\'' ) ) )* '\''
    	;

INTEGER	
	:	('0'..'9')+
	;

FLOAT
	:	('0'..'9')* '.' ('0'..'9')+
	;

function
	:	FUNC '(' ( logicalExpression (',' logicalExpression)* )? ')' -> ^(FUNC logicalExpression*)
	;

FUNC
	:	('a'..'z' | 'A'..'Z' | '_' ) ('a'..'z' | 'A'..'Z' | '0'..'9' | '_' )*
	;

fragment EscapeSequence 
	:	'\\'
  	(	
  		'n' 
	|	'r' 
	|	't'
	|	'\'' 
	|	'\\'
	|	UnicodeEscape
	)
  ;

fragment UnicodeEscape
    	:    	'u' HexDigit HexDigit HexDigit HexDigit 
    	;

fragment HexDigit 
	: 	('0'..'9'|'a'..'f'|'A'..'F') ;

WS	
	:  (' '|'\r'|'\t'|'\u000C'|'\n') {$channel=Hidden;}
	;